import os
import glob
from unitgrade_private import load_token
from cs103.report3_complete import Report3
from unitgrade_private.hidden_create_files import setup_grade_file_report
from unitgrade_private.deployment import remove_hidden_methods
from unitgrade_private.docker_helpers import docker_run_token_file
from unitgrade_private.docker_helpers import download_docker_images, compile_docker_image
from snipper.snip_dir import snip_dir

if __name__ == "__main__": #!s=docker_a
    # Step 1: Deploy the students files and return the directory they were written to
    setup_grade_file_report(Report3)  # Create report3_complete_grade.py which tests everything

    fout, Report = remove_hidden_methods(Report3, outfile="report3.py")  # Create report3.py without @hide-methods
    setup_grade_file_report(Report)                                      # Create report3_grade.py for the students

    student_directory = "../../students/cs103"
    snip_dir("./", student_directory, exclude=['*.token', 'deploy.py', 'report3_complete*.py', '.*']) #!s

    # Step 2: Simulate that the student run their report script and generate a .token file.
    os.system("cd ../../students && python -m cs103.report3_grade") #!s=docker_b
    student_token_file = glob.glob(student_directory + "/*.token").pop()  #!s

    # Step 3: Compile the Docker image (obviously you should only do this once). #!s=docker_c
    download_docker_images(destination="../docker")  # Download an up-to-date docker image from gitlab.
    Dockerfile = "../docker/unitgrade-docker/Dockerfile" # Location of just downloaded docker file
    compile_docker_image(Dockerfile, tag="unitgrade-docker") #!s

    # Step 4: Test the students code in the .token file and get the results-token-file:  #!s=docker_d
    token = docker_run_token_file(Dockerfile_location=Dockerfile,
                                  host_tmp_dir=os.path.dirname(Dockerfile) + "/home",
                                  student_token_file=student_token_file,
                                  instructor_grade_script="report3_complete_grade.py",
                                  tag="unitgrade-docker") #!s

    # Load the two token files and compare their scores #!s=docker_e
    checked_token, _ = load_token(token)
    results, _ = load_token(student_token_file)

    print("Student's score was:", results['total'])
    print("My independent evaluation of the students score was", checked_token['total']) #!s

    # Save it for the Readme
    s = ""
    s += f"Student's score was: {results['total']}\n"
    s += f"My independent evaluation of the students score was {checked_token['total']}"
    with open("docker_results.txt", 'w') as f:
        f.write(s)
