class Fraction: #!s
    def __init__(self, n, m):
        self.n = n
        self.m = m

    def __add__(self, other): #!f f should be the result of adding fraction 'other' to this fraction.
        """ Computes f = f1 + f2 (where f1 and f2 are both Fraction-objects and f1=self, f2=other), see \cref{eq:fraction} and \cite{aa203} """ #!s
        f = Fraction(other.m * self.n + other.n * self.m, self.n * self.m)
        return f #!s

    def __mul__(self, other): #!f f should be the result of multiplying fraction 'other' to this fraction.
        # Overwrite to implement f = f1 * f2 #!s
        f = Fraction(other.n * self.n, other.m*self.m)
        return f #!s

    def __truediv__(self, other): #!f f should be the result of dividing this fraction with 'other'
        # Overwrite to implement f = f1/f2, or more specifically self/other. #!s
        f = self.__mul__(Fraction(other.m, other.n))
        return f #!s

    def __str__(self):
        """ Creates a string representation. You can use it as `print(str(Fraction(1,2))) to output 1/2"""
        return f"{self.n} / {self.m}" #!s


def from_string(s):
    """ Convert the string s to a Fraction(a, b) object. """
    if '+' in s: #!b
        l = [from_string(ss.strip()) for ss in s.split("+")]
        return l[0] + l[1]
    if 'div' in s:
        l = [from_string(ss.strip()) for ss in s.split("div")]
        return l[0] / l[1]
    if '*' in s:
        l = [from_string(ss.strip()) for ss in s.split("*")]
        return l[0] * l[1]
    if '/' in s:
        return Fraction(*[int(ss.strip()) for ss in s.split("/")]) #!b Compute and return a Fraction(a,b) object corresponding to s.


if __name__ == "__main__": #!o=a
    f1 = Fraction(1, 2)  # Represents 1/2 #!s=a
    f2 = Fraction(3, 5) # Represents 3/5
    print(f"Result of {f1} + {f2} is", f1+f2) #!s=a
    #!o=a
    # Now do some compound tests:
    s = " 1 / 4 * 1 / 2" #!s=b #!o=b
    print("Result of", s, "is", from_string(s))
    s =  "5 / 2 div 10 / 3"
    print("Result of", s, "is", from_string(s)) #!s=b #!o=b

