from unitgrade import UTestCase, Report
import cpp_exam
from cpp_exam.problem1_vectors import Vector, createField, displayField, scaleField, addFields
from cpp_exam.problem2_rle import Elem, rle_sum, reverse, concatenate, displayRLEList
from cpp_exam.problem3_groceries import GroceryList
from cpp_exam.problem4_filter import Buffer


def vector2string(v):
    return f"({v.x},{v.y})"

def check_field(self, A):
    print("\nChecking size and type of field...")
    self.assertIsInstance(A, list)
    self.assertIsInstance(A[0], list)
    self.assertIsInstance(A[0][0], Vector)
    print("Checking all elements of field...")
    for i, r in enumerate(A):
        for j, a in enumerate(r):
            # if i > 3 or j > 3: continue
            print(f"Checking that A[{i}][{j}] = {self.get_expected_test_value()};", "your value was", vector2string(A[i][j]))
            self.assertEqualC(vector2string(A[i][j]))

class Q1Vectors(UTestCase):
    def test_correct_format(self):
        """ Test your field is in the right format.
            Hints:
                * If this test fails, all subsequent tests will fail. Make sure it works!
        """
        A = createField(3, 3, Vector(1, 2))
        self.assertEqual(len(A), 3) # Check there are 3 rows in A
        self.assertEqual(len(A[0]), 3)  # Check there are 3 columns in A
        v = A[1][1]                 # Should be a Vector(1, 2) element.
        self.assertEqual(v.x, 1)  # Check that first coordinate is 1
        self.assertEqual(v.y, 2)  # Check that first coordinate is 1

    def test_createField_small(self):
        # Test a 4x4 field
        check_field(self, createField(4, 4, Vector(1,2)))
        # Test an 8x8 field
        check_field(self, createField(8, 8, Vector(3,8)))

    def test_display_field(self):
        A = createField(4, 3, Vector(1,2))
        A[0][0] = A[1][1] = A[2][2] = Vector(0,0)
        self.assertEqualC(displayField(A))

        B = createField(6,8, Vector(3, -9))
        B[0][1] = B[2][1] = B[3][2] = B[4][5] = B[5][6] = Vector(0,0)
        self.assertEqualC(displayField(B))

    def test_add_fields(self):
        A = createField(4, 3, Vector(1,2))
        B = createField(4, 3, Vector(3, -9))
        check_field(self, addFields(A, B))

    def test_scale_fields(self):
        A = createField(2, 4, Vector(1,2))
        check_field(self, scaleField(A, 2))
        A = createField(6, 6, Vector(3, 1))
        check_field(self, scaleField(A, 3))

def make_rle1():
    return Elem(4, 10, Elem(12, 6, Elem(10, 5, None)))

def make_rle2():
    return Elem(4, 2, Elem(5, 3, Elem(3, 5, None)))

def make_rle3():
    return Elem(6, 3, Elem(7, 5, Elem(8, 5, None)))

class Q2RLE(UTestCase):
    def test_reverse_empty(self):
        self.assertEqual(reverse(None), None)

    def test_reverse(self):
        self.assertEqualC(displayRLEList(reverse(make_rle1())))
        l = make_rle1() # Test reversal of the tail
        self.assertEqualC(displayRLEList(reverse(l.next)))

    def test_concatenate_with_empty(self):
        l = make_rle1()
        print(f"Concatenating {displayRLEList(l)} with empty list")
        self.assertEqualC(displayRLEList(concatenate(l, None)))

        l = make_rle2()
        print(f"Concatenating {displayRLEList(l)} with empty list")
        self.assertEqualC(displayRLEList(concatenate(l, None)))

    def test_concatenate_with_another(self):
        l1, l2 = make_rle1(), make_rle2()
        print(f"Concatenating {displayRLEList(l1)} with {displayRLEList(l2)}")
        self.assertEqualC(displayRLEList(concatenate(l1, l2)))

    def test_concatenate_three_lists(self):
        l1, l2, l3 = make_rle1(), make_rle2(), make_rle3()
        print(f"Concatenating {displayRLEList(l1)} with {displayRLEList(l2)} and {displayRLEList(l3)}")
        self.assertEqualC(displayRLEList(concatenate(l1, concatenate(l2, l3))))

    def test_sum_empty(self):
        print(f"Computing sum of empty list")
        self.assertEqual(rle_sum(None), 0)

    def test_sum_nonempty(self):
        for l in [ make_rle1(),  make_rle2(),  make_rle3()]:
            print(f"Computing sum of {displayRLEList(l)}")
            self.assertEqualC(rle_sum(l), 0)

class Q3Groceries(UTestCase):
    def test_add(self):
        gl = GroceryList()
        print("Initial list")
        gl.display()
        print("After adding cheddar:")
        gl.add("Cheddar", 500, "Not too mature")
        gl.display()
        print("After adding more spinach:")
        gl.add("Spinach", 200, "Baby spinach if available")
        gl.display()
        print("After adding even more spinach:")
        gl.add("Spinach", 200, "Frozen is OK")
        gl.display()

    def test_remove1(self):
        gl = GroceryList()
        gl.display()
        print("After removing spinach")
        self.assertTrue(gl.remove("Spinach", 200))
        self.assertEqualC(gl.display())

        print("After removing more spinach")
        self.assertTrue(gl.remove("Spinach", 100))
        self.assertEqualC(gl.display())

        print("Trying to remove even more spinach:")
        self.assertFalse(gl.remove("Spinach", 100))

        print("Trying to remove cheddar:")
        self.assertFalse(gl.remove("Cheddar", 100))

    def test_copy(self):
        gl = GroceryList()
        print("Initial grocery list:")
        gl.display()
        print("After copying spinach into baby spinach:")
        self.assertTrue(gl.copyEntry("Spinach", "Baby spinach"))
        self.assertEqualC(gl.display())
        print("After copying baby spinach into dessert:")
        self.assertFalse(gl.copyEntry("Baby spinach", "Dessert"))
        print("After copying Cheddar into Spinach")
        self.assertFalse( gl.copyEntry("Cheddar", "Spinach"))

class Q4FilterBuffer(UTestCase):
    def test_if_compiles(self):
        b = Buffer()
        b.write(42)
        b.write(42)
        b.occupancy()
        b.read()
        print("None of the functions crashed! Congrats!")

    def test_occupancy(self):
        b = Buffer()
        print("Current occupancy", b.occupancy())
        self.assertEqual(b.occupancy(), 0)
        for i in range(7):
            b.write(i*10)
        print("Current buffer occupancy: ", b.occupancy())
        self.assertEqual(b.occupancy(), 7)
        for i in range(5, 10):
            b.write(i*10)
        print("Current buffer occupancy: ", b.occupancy())
        self.assertEqual(b.occupancy(), 10)

    def test_read(self):
        b = Buffer()
        print("Current buffer occupancy: ", b.occupancy())
        print(f"Reading from empty buffer should return default value of {b.default}")
        self.assertEqual(b.read(), -999)

        for i in range(5):
            b.write(i*10)

        print("Current buffer occupancy: ", b.occupancy() )
        self.assertEqual(b.occupancy(), 5)
        for i in range(3):
            r = b.read()
            print("Reading from the buffer returns: ", r)
            self.assertEqual(r, i*10)
        print("Current buffer occupancy: ", b.occupancy() )
        self.assertEqualC(b.occupancy())

        for i in range(8):
            b.write(i * 10)
        print("Current buffer occupancy: ", b.occupancy())
        self.assertEqualC(b.occupancy())

        for i in range(6):
            r = b.read()
            print("Reading from the buffer returns: ", r)
            self.assertEqualC(r)

    def test_reset(self):
        b = Buffer()

        for i in range(5):
            b.write(i*10)
        print("Current buffer occupancy: ", b.occupancy() )
        self.assertEqual(b.occupancy(), 5)
        for i in range(3):
            r = b.read()
            print("Reading from the buffer returns: ", r)
            self.assertEqual(r, i*10)
        print("Current buffer occupancy: ", b.occupancy() )
        self.assertEqualC(b.occupancy())
        b.reset()
        for i in range(8):
            b.write(i * 10)
        print("Current buffer occupancy: ", b.occupancy())
        self.assertEqualC(b.occupancy())

        for i in range(3):
            r = b.read()
            print("Reading from the buffer returns: ", r)
            self.assertEqualC(r)

class ExamMay2021(Report):
    title = "Programming in C++: Exam May 2021"
    pack_imports = [cpp_exam]
    questions = [(Q1Vectors, 25), (Q2RLE, 25), (Q3Groceries, 25), (Q4FilterBuffer, 25)]

if __name__ == "__main__":
    from unitgrade import evaluate_report_student
    evaluate_report_student(ExamMay2021())
