import numpy as np
from unitgrade import UTestCase, Report, hide
import intro_python
from intro_python.problems import water_height, tictactoe, time_angle, astronomical_season, standardize_address

class Q1_WaterHeight(UTestCase):
    def test1(self):
        h0 = 5
        r = np.array([4.5, 0, 1.5, 0, 0, 0.5, 1, 2, 5])
        h = water_height(h0, r)
        print("Water height computed to be", h, "should be", 3.0)
        self.assertEqual(h, 3.0) # Check that the height is 3.0

    @hide
    def test_water_height_hidden(self):
        checks = [(120, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                  (12, []), (14.2, [8.8]), (0, [0.8]),
                  (3, [0, 1, 2, 0, 0, 5.2, 6.5, 7.1, 0, 0, 0.1, 1, 2.9, 0.13, 0, 2, 0, 0]),
                  (0, [0, 5, 2, 0, 0, 5.2, 6.5, 7.1, 0, 0, 0.1, 1, 2.9, 0.13, 0, 1.2, 0, 1.2, 7.5, 0]),
                  (0, [0, 0, 2, 2.1, 2.4, 2.2, 2.5]), (18, [30, 1, 28.8]), (1, [0.5]), (2, [])]
        for h0, r in checks:
            self.assertEqualC(water_height(h0, r))

class Q2_AstronomicalSeason(UTestCase):
    def test_seasons(self):
        season = astronomical_season('09/12-2020')
        print("Season was computed to be", season, "it was supposed to be", self.get_expected_test_value())
        self.assertEqualC(season)

    @hide
    def test_seasons_hidden(self):
        for d in ['27/12-1998', '21/06-2108', '08/05-1998', '07/08-1945', '22/12-1208', '19/03-2001', '23/09-2018', '21/06-2008','12/04-1964', '13/01-1900']:
            self.assertEqualC(astronomical_season(d))


class Q3_TimeAngle(UTestCase):
    def test_angle(self):
        a = time_angle(hour=8, minute=20)
        print("Angle was", a, "it was supposed to be", self.get_expected_test_value())
        self.assertEqualC(a)

    @hide
    def test_angle_extended(self):
        for minute in [0, 15, 18, 20, 34, 50, 59]:
            for hour in [0, 1, 5, 6, 10, 12]:
                self.assertEqualC(time_angle(hour, minute))

class Q4_TicTacToe(UTestCase):
    def test_tic_tac(self):
        board = np.array([[2, 1, 1],
                          [1, 1, 2],
                          [2, 0, 0]])
        score = tictactoe(board)
        print("Score for board was", score, "it is supposed to be", self.get_expected_test_value())
        self.assertEqualC(score)

    @hide
    def test_tic_tac_hidden(self):
        boards = [[[1, 2, 0], [1, 2, 0], [1, 2, 0]], [[1, 1, 1], [2, 1, 2], [2, 2, 1]],
                  [[2, 0, 1], [2, 1, 0], [0, 0, 2]], [[1, 0, 2], [0, 1, 0], [2, 0, 1]],
                  [[2, 0, 1], [0, 2, 1], [0, 0, 1]], [[0, 1, 0], [0, 1, 1], [2, 2, 2]],
                  [[1, 1, 2], [0, 2, 0], [2, 1, 0]], [[1, 1, 1], [0, 2, 0], [0, 0, 0]],
                  [[1, 2, 1], [2, 1, 0], [2, 0, 1]], [[0, 0, 0], [0, 1, 0], [0, 0, 0]]]
        for board in boards:
            self.assertEqualC(tictactoe(np.asarray(board)))


class Q5_StandardizeAddress(UTestCase):
    def test_standardize_address(self):
        s = standardize_address('New York 10001')
        print("Address computed to be", s, "was supposed to be", self.get_expected_test_value())
        self.assertEqualC(s)

    @hide
    def test_standardize_address_hidden(self):
        for address in ['Kongens_Lyngby_2800', '10000_Zagreb','43500 Daruvar','Egtved_6040','Vejle 7200', '02108_Boston',
                  'Pasadena_91001', '90001_Los_Angeles', 'San_Francisco_94016', 'Rio_de_Jainero_22775']:
            self.assertEqualC(standardize_address(address))



class Exam2021(Report):
    title = "Introduction to Python: Exam spring 2021"
    pack_imports = [intro_python]
    questions = [(Q1_WaterHeight, 20),
                 (Q2_AstronomicalSeason, 20),
                 (Q3_TimeAngle, 20),
                 (Q4_TicTacToe, 20),
                 (Q5_StandardizeAddress, 20)]

if __name__ == "__main__":
    from unitgrade import evaluate_report_student
    evaluate_report_student(Exam2021())
